import os
import shutil
import tarfile
import tempfile
import zlib

from keyczar.errors import KeyczarError
from keyczar.errors import InvalidSignatureError

from sesame.utils import ask_overwrite
from sesame.utils import make_secure_temp_directory
from sesame.utils import mkdir_p


def encrypt(inputfiles, outputfile, keys):
    with make_secure_temp_directory() as working_dir:
        # create a tarfile of inputfiles
        with tarfile.open(os.path.join(working_dir, 'sesame.tar'), 'w') as tar:
            for name in inputfiles:
                # TODO use warning to prompt user here
                if os.path.isabs(name):
                    # fix absolute paths, same as tar does
                    tar.add(name, arcname=name[1:])
                elif name.startswith('..'):
                    # skip relative paths
                    pass
                else:
                    tar.add(name)

        try:
            # encrypt the tarfile
            with open(os.path.join(working_dir, 'sesame.tar'), 'rb') as i:
                with open(outputfile, 'wb') as o:
                    o.write(keys[0].Encrypt(zlib.compress(i.read())))
        except KeyczarError as e:
            raise SesameError(
                'An error occurred in keyczar.Encrypt\n  {0}:{1}'.format(e.__class__.__name__, e)
            )


def decrypt(inputfile, keys, force=False, output_dir=None, try_all=False):
    with make_secure_temp_directory() as working_dir:
        # create a temporary file
        working_file = tempfile.mkstemp(dir=working_dir)

        success = False

        # iterate all keys; first successful key will break
        for key in keys:
            try:
                # decrypt the input file
                with open(inputfile, 'rb') as i:
                    data = zlib.decompress(key.Decrypt(i.read()))

                # write into our working file
                with os.fdopen(working_file[0], 'wb') as o:
                    o.write(data)

                success = True
                break
            except InvalidSignatureError as e:
                if try_all is False:
                    raise SesameError('Incorrect key')
            except KeyczarError as e:
                raise SesameError(
                    'An error occurred in keyczar.Decrypt\n  {0}:{1}'.format(
                        e.__class__.__name__, e
                    )
                )

        # check failure
        if success is False:
            raise SesameError('No valid keys for decryption')

        if tarfile.is_tarfile(working_file[1]):
            # untar the decrypted temp file
            with tarfile.open(working_file[1], 'r') as tar:
                tar.extractall(path=working_dir)

            # get list of full paths to all files in working dir
            # ignoring the decrypted tarfile
            working_items = [
                os.path.join(rootdir[len(working_dir)+1:], filename)
                for rootdir, dirnames, filenames in os.walk(working_dir)
                for filename in filenames if filename != os.path.basename(working_file[1])
            ]

            for filename in working_items:
                # move all files to the output path
                move_output_file(
                    path=os.path.join(working_dir, filename),
                    dest=os.path.join(output_dir, filename),
                    force=force,
                )

        else:
            # older versions of Sesame didn't wrap a tarfile and
            # encrypted only a single file at a time
            working_items = [os.path.join(working_dir[len(working_dir)+1:], )]

            # attempt to create an output filename using old Sesame logic
            if inputfile.endswith(".encrypted"):
                move_output_file(
                    path=working_file[1],
                    dest=os.path.join(output_dir, inputfile[0:-10]),
                    force=force,
                )
            else:
                # create a secure random-named file
                with tempfile.NamedTemporaryFile(suffix='.sesame-decrypted', dir=output_dir, delete=False) as keyfile:
                    keyfile.write("\0")

                # overwrite the file just created with the decrypted file
                move_output_file(
                    path=working_file[1],
                    dest=os.path.join(output_dir, keyfile.name),
                    force=True,
                )


def move_output_file(path, dest, force=False):
    # ask user about overwrite
    if force is False and os.path.exists(dest):
        if ask_overwrite(dest) is False:
            return

    # ensure destination dirs exist
    mkdir_p(os.path.dirname(dest))

    # move file to output_dir
    shutil.copy(path, dest)


class SesameError(Exception):
    pass



#! /usr/bin/env python

import os
import sys

import sesame

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

packages = [
    'sesame',
]

requires = [
    open('requirements.txt').read()
]

if sys.version_info < (2, 7):
    requires += ['argparse']

setup(
    name='sesame',
    version=sesame.__version__,
    description='Encryption tool for application configs.',
    long_description=open('README.rst').read(),
    author='Matt Black',
    author_email='dev@mafro.net',
    url='http://github.com/mafrosis/sesame',
    packages=packages,
    package_data={'': ['LICENSE']},
    package_dir={'': '.'},
    include_package_data=True,
    install_requires=requires,
    scripts=['scripts/sesame'],
    license=open('LICENSE').read(),
    classifiers=(
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Natural Language :: English',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ),
)



import argparse
import os
import sys

from sesame.core import decrypt
from sesame.core import encrypt
from sesame.core import SesameError

from sesame.utils import ask_create_key
from sesame.utils import ask_overwrite
from sesame.utils import confirm
from sesame.utils import find_sesame_keys
from sesame.utils import read_key

MODE_ENCRYPT = 1
MODE_DECRYPT = 2


def entrypoint():
    try:
        # setup and run argparse
        args = parse_command_line()

        # locate encryption keys
        keys = get_keys(args)

        # check we have a key
        if len(keys) == 0:
            raise SesameError('No keys provided')

        # run encrypt/decrypt
        main(args, keys)

    except SesameError as e:
        sys.stderr.write('{0}\n'.format(e))
        sys.exit(1)


def parse_command_line():
    parser = argparse.ArgumentParser(
        description='Sesame config file encryption and decryption'
    )
    subparsers = parser.add_subparsers()

    # setup the arguments for both encrypt and decrypt
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument(
        '-k', '--keyfile',
        help='Path to keyczar encryption key')

    # setup parser for encrypt command
    pencrypt = subparsers.add_parser('encrypt',
        parents=[parent_parser],
        help='Encrypt one or more files',
    )
    pencrypt.set_defaults(mode=MODE_ENCRYPT)
    pencrypt.add_argument(
        'outputfile',
        help='Encrypted file to be created')
    pencrypt.add_argument(
        'inputfile', nargs='+',
        help='Files to be encrypted')
    pencrypt.add_argument(
        '-f', '--force', action='store_true',
        help='Force overwrite of existing encrypted file')

    # setup parser for decrypt command
    pdecrypt = subparsers.add_parser('decrypt',
        parents=[parent_parser],
        help='Decrypt a file created with Sesame',
    )
    pdecrypt.set_defaults(mode=MODE_DECRYPT)
    pdecrypt.add_argument(
        'inputfile',
        help='File to be decrypted')
    pdecrypt.add_argument(
        '-f', '--force', action='store_true',
        help='Force overwrite of existing decrypted file')
    pdecrypt.add_argument(
        '-O', '--output-dir', default=os.getcwd(),
        help='Extract files into a specific directory')
    pdecrypt.add_argument(
        '-T', '--try-all', action='store_true',
        help='Search for keys from current directory and try all of them')

    return parser.parse_args()


def get_keys(args):
    """
    Get the set of keys to be used for this encrypt/decrypt
    """
    keys = []
    if args.keyfile is None:
        # attempt to locate a key
        keys = find_sesame_keys()

        if len(keys) == 0:
            # ask the user to generate one
            key = ask_create_key()
            if key is not None:
                keys = [key]

        elif len(keys) >= 1:
            if args.mode == MODE_ENCRYPT or (args.mode == MODE_DECRYPT and args.try_all is False):
                # ask the user if they want to use the first key found
                if confirm('No key supplied and {0} found. Use {1}?'.format(
                    len(keys), keys.keys()[0]
                ), default=True):
                    keys = [keys.items()[0][1]]
                else:
                    # ask the user to generate one
                    key = ask_create_key()
                    if key is not None:
                        keys = [key]
                    else:
                        return
            else:
                # --try-all flag was supplied
                keys = keys.values()
        else:
            # use the only key found
            keys = [keys.items()[0][1]]
    else:
        # create a key
        if os.path.exists(args.keyfile) is False:
            key = ask_create_key()
            if key is not None:
                keys = [key]
        else:
            # load the single key supplied
            key = read_key(os.path.join(args.keyfile))
            if key is not None:
                keys = [key]

    return keys


def main(args, keys):
    if args.mode == MODE_ENCRYPT:
        # check if destination exists
        if args.force is False and os.path.exists(args.outputfile):
            if ask_overwrite(args.outputfile) is False:
                return

        # args.inputfile is a list of files
        for f in args.inputfile:
            if not os.path.exists(f):
                raise SesameError('File doesn\'t exist at {0}'.format(f))

        encrypt(
            inputfiles=args.inputfile,
            outputfile=args.outputfile,
            keys=keys
        )

    elif args.mode == MODE_DECRYPT:
        # fail if input file doesn't exist
        if not os.path.exists(args.inputfile):
            raise SesameError('File doesn\'t exist at {0}'.format(args.inputfile))

        # check input not zero-length
        statinfo = os.stat(args.inputfile)
        if statinfo.st_size == 0:
            raise SesameError('Input file is zero-length')

        decrypt(
            inputfile=args.inputfile,
            keys=keys,
            force=args.force,
            output_dir=args.output_dir,
            try_all=args.try_all
        )



import collections
import contextlib
import errno
import fnmatch
import os
import shutil
import tempfile

from keyczar.keys import AesKey


def find_sesame_keys():
    # use OrderedDict to maintain order in which keys are found
    keys = collections.OrderedDict()
    for root, dirs, files in os.walk(os.getcwd()):
        for filename in fnmatch.filter(files, '*.key'):
            try:
                keys[filename] = read_key(os.path.join(root, filename))
            except (ValueError, KeyError):
                pass
    return keys


def read_key(key_path):
    with open(key_path, 'r') as f:
        data = f.read()
    return AesKey.Read(data)


def ask_create_key(directory=os.getcwd()):
    res = raw_input('Encryption key not provided. Create? [Y/n] ')
    if len(res) > 0 and not res.lower().startswith('y'):
        return None

    return create_key(directory)


def create_key(directory, write=True):
    # generate a new key
    key = AesKey.Generate()

    if write is True:
        # create a unique file to house our generated key
        with tempfile.NamedTemporaryFile(prefix='sesame', suffix='.key', dir=directory, delete=False) as keyfile:
            keyfile.write(str(key))

        print 'Encryption key created at {0}'.format(os.path.basename(keyfile.name))

    return key


def ask_overwrite(path, isdir=False):
    noun = 'Directory' if os.path.isdir(path) else 'File'
    return confirm(
        '{0} {1} exists. Overwrite?'.format(noun, os.path.basename(path)),
        default=False
    )


def ask_decrypt_file(filename):
    return confirm('Found {0}. Do you want to decrypt this file?'.format(filename), default=False)


def confirm(msg, default=True):
    """
    Display confirm prompt on command line

    msg:
        The message to display to the user
    default:
        Default True/False (yes/no) at the prompt
    """
    if default is True:
        display = '{0} [Y/n] '.format(msg)
    else:
        display = '{0} [y/N] '.format(msg)

    res = raw_input(display)
    if default is True:
        return True if len(res) == 0 or res.lower().startswith('y') else False
    else:
        return True if len(res) > 0 and res.lower().startswith('y') else False


@contextlib.contextmanager
def make_secure_temp_directory():
    temp_dir = tempfile.mkdtemp()
    try:
        yield temp_dir
    except Exception as e:
        raise e
    finally:
        shutil.rmtree(temp_dir)


def mkdir_p(path):
    """ http://stackoverflow.com/a/600612/425050 """
    try:
        os.makedirs(path)
    except OSError as exc: # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else: raise
